import asyncio

from asynctest import TestCase, CoroutineMock

from atlassian_jwt_auth.contrib.aiohttp import (
    JWTAuthVerifier, HTTPSPublicKeyRetriever)
from atlassian_jwt_auth.tests import utils, test_verifier


class SyncJWTAuthVerifier(JWTAuthVerifier):

    def __init__(self, *args, loop=None, **kwargs):
        if loop is None:
            loop = asyncio.get_event_loop()
        self.loop = loop
        super().__init__(*args, **kwargs)

    def verify_jwt(self, *args, **kwargs):
        return self.loop.run_until_complete(
            super().verify_jwt(*args, **kwargs)
        )


class JWTAuthVerifierTestMixin(test_verifier.BaseJWTAuthVerifierTest):
    loop = None

    def _setup_mock_public_key_retriever(self, pub_key_pem):
        m_public_key_ret = CoroutineMock(spec=HTTPSPublicKeyRetriever)
        m_public_key_ret.retrieve.return_value = pub_key_pem.decode()
        return m_public_key_ret

    def _setup_jwt_auth_verifier(self, pub_key_pem, **kwargs):
        m_public_key_ret = self._setup_mock_public_key_retriever(pub_key_pem)
        return SyncJWTAuthVerifier(m_public_key_ret, loop=self.loop, **kwargs)


class JWTAuthVerifierRS256Test(
        utils.RS256KeyTestMixin, JWTAuthVerifierTestMixin, TestCase):
    """Tests for aiohttp.JWTAuthVerifier class for RS256 algorithm"""


class JWTAuthVerifierES256Test(
        utils.ES256KeyTestMixin, JWTAuthVerifierTestMixin, TestCase):
    """Tests for aiohttp.JWTAuthVerifier class for ES256 algorithm"""



import base64
import unittest

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.hazmat.primitives import serialization

from atlassian_jwt_auth.signer import JWTAuthSigner
from atlassian_jwt_auth.tests import utils
from atlassian_jwt_auth.key import DataUriPrivateKeyRetriever


def convert_key_pem_format_to_der_format(private_key_pem):
    private_key = load_pem_private_key(private_key_pem,
                                       password=None,
                                       backend=default_backend())
    return private_key.private_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption()
    )


class BaseDataUriPrivateKeyRetrieverTest(object):
    """ tests for the DataUriPrivateKeyRetriever class. """

    def setUp(self):
        self._private_key_pem = self.get_new_private_key_in_pem_format()
        self._public_key_pem = utils.get_public_key_pem_for_private_key_pem(
            self._private_key_pem)
        self._private_key_der = convert_key_pem_format_to_der_format(
            self._private_key_pem)

    def get_example_data_uri(self, private_key_der):
        return ('data:application/pkcs8;kid=example%2Feg;base64,' +
                base64.b64encode(private_key_der).decode('utf-8'))

    def test_load_data_uri(self):
        """ tests that a valid data uri is correctly loaded. """
        expected_kid = 'example/eg'
        data_uri = self.get_example_data_uri(self._private_key_der)
        provider = DataUriPrivateKeyRetriever(data_uri)
        kid, private_key_pem = provider.load('example')
        self.assertEqual(kid.key_id, expected_kid)
        self.assertEqual(private_key_pem,
                         self._private_key_pem.decode('utf-8'))

    def test_load_data_uri_can_be_used_with_a_signer(self):
        """ tests that the data uri private key retriever can be used with a
            signer to generate a jwt.
        """
        data_uri = self.get_example_data_uri(self._private_key_der)
        provider = DataUriPrivateKeyRetriever(data_uri)
        jwt_auth_signer = JWTAuthSigner(
            'issuer', provider, algorithm=self.algorithm)
        jwt_auth_signer.generate_jwt('aud')


class DataUriPrivateKeyRetrieverRS256Test(BaseDataUriPrivateKeyRetrieverTest,
                                          utils.RS256KeyTestMixin,
                                          unittest.TestCase):
    pass


class DataUriPrivateKeyRetrieverES256Test(BaseDataUriPrivateKeyRetrieverTest,
                                          utils.ES256KeyTestMixin,
                                          unittest.TestCase):
    pass



import calendar
import datetime
import random

import jwt
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from atlassian_jwt_auth import algorithms
from atlassian_jwt_auth import key


class JWTAuthSigner(object):

    def __init__(self, issuer, private_key_retriever, **kwargs):
        self.issuer = issuer
        self.private_key_retriever = private_key_retriever
        self.lifetime = kwargs.get('lifetime', datetime.timedelta(hours=1))
        self.algorithm = kwargs.get('algorithm', 'RS256')
        self.subject = kwargs.get('subject', None)
        self._private_keys_cache = dict()

        if self.algorithm not in set(
                algorithms.get_permitted_algorithm_names()):
            raise ValueError("Algorithm, '%s', is not permitted." %
                             self.algorithm)
        if self.lifetime > datetime.timedelta(hours=1):
            raise ValueError("lifetime, '%s',exceeds the allowed 1 hour max" %
                             (self.lifetime))

    def _obtain_private_key(self, key_identifier, private_key_pem):
        """ returns a loaded instance of the given private key either from
            cache or from the given private_key_pem.
        """
        priv_key = self._private_keys_cache.get(key_identifier.key_id, None)
        if priv_key is not None:
            return priv_key
        if not isinstance(private_key_pem, bytes):
            private_key_pem = private_key_pem.encode()
        priv_key = serialization.load_pem_private_key(
            private_key_pem,
            password=None,
            backend=default_backend()
        )
        if len(self._private_keys_cache) > 10:
            self._private_keys_cache = dict()
        self._private_keys_cache[key_identifier.key_id] = priv_key
        return priv_key

    def _generate_claims(self, audience, **kwargs):
        """ returns a new dictionary of claims. """
        now = self._now()
        claims = {
            'iss': self.issuer,
            'exp': now + self.lifetime,
            'iat': now,
            'aud': audience,
            'jti': '%s:%s' % (
                now.strftime('%s'), random.SystemRandom().getrandbits(32)),
            'nbf': now,
            'sub': self.subject or self.issuer,
        }
        claims.update(kwargs.get('additional_claims', {}))
        return claims

    def _now(self):
        return datetime.datetime.utcnow()

    def generate_jwt(self, audience, **kwargs):
        """ returns a new signed jwt for use. """
        key_identifier, private_key_pem = self.private_key_retriever.load(
            self.issuer)
        private_key = self._obtain_private_key(
            key_identifier, private_key_pem)
        token = jwt.encode(
            self._generate_claims(audience, **kwargs),
            key=private_key,
            algorithm=self.algorithm,
            headers={'kid': key_identifier.key_id})
        if isinstance(token, str):
            token = token.encode('utf-8')
        return token


class TokenReusingJWTAuthSigner(JWTAuthSigner):

    def __init__(self, issuer, private_key_retriever, **kwargs):
        super(TokenReusingJWTAuthSigner, self).__init__(
            issuer, private_key_retriever, **kwargs)
        self.reuse_threshold = kwargs.get('reuse_jwt_threshold', 0.95)

    def get_cached_token(self, audience, **kwargs):
        """ returns the cached token. If there is no matching cached token
            then None is returned.
        """
        return getattr(self, '_previous_token', None)

    def set_cached_token(self, value):
        """ sets the cached token."""
        self._previous_token = value

    def can_reuse_token(self, existing_token, claims):
        """ returns True if the provided existing token can be reused
            for the claims provided.
        """
        if existing_token is None:
            return False
        existing_claims = jwt.decode(
            existing_token, verify=False, options={'verify_signature': False})
        existing_lifetime = (int(existing_claims['exp']) -
                             int(existing_claims['iat']))
        this_lifetime = (claims['exp'] - claims['iat']).total_seconds()
        if existing_lifetime != this_lifetime:
            return False
        about_to_expire = int(existing_claims['iat']) + (
            self.reuse_threshold * existing_lifetime)
        if calendar.timegm(self._now().utctimetuple()) > about_to_expire:
            return False
        if set(claims.keys()) != set(existing_claims.keys()):
            return False
        for dict_key, val in claims.items():
            if dict_key in ['exp', 'iat', 'jti', 'nbf']:
                continue
            if existing_claims[dict_key] != val:
                return False
        return True

    def generate_jwt(self, audience, **kwargs):
        existing_token = self.get_cached_token(audience, **kwargs)
        claims = self._generate_claims(audience, **kwargs)
        if existing_token and self.can_reuse_token(existing_token, claims):
            return existing_token
        token = super(TokenReusingJWTAuthSigner, self).generate_jwt(
            audience, **kwargs)
        self.set_cached_token(token)
        return token


def _create_signer(issuer, private_key_retriever, **kwargs):
    signer_cls = JWTAuthSigner
    if kwargs.get('reuse_jwts', None):
        signer_cls = TokenReusingJWTAuthSigner
    return signer_cls(issuer, private_key_retriever, **kwargs)


def create_signer(issuer, key_identifier, private_key_pem, **kwargs):
    private_key_retriever = key.StaticPrivateKeyRetriever(
        key_identifier, private_key_pem)
    return _create_signer(issuer, private_key_retriever, **kwargs)


def create_signer_from_file_private_key_repository(
        issuer, private_key_repository, **kwargs):
    private_key_retriever = key.FilePrivateKeyRetriever(private_key_repository)
    return _create_signer(issuer, private_key_retriever, **kwargs)



from functools import wraps
from jwt.exceptions import InvalidIssuerError, InvalidTokenError

from .asap import _process_asap_token, _verify_issuers
from .utils import SettingsDict


def _with_asap(func=None, backend=None, issuers=None, required=True,
               subject_should_match_issuer=None):
    if backend is None:
        raise ValueError(
            'Invalid value for backend. Use a subclass instead.'
        )

    def with_asap_decorator(func):
        @wraps(func)
        def with_asap_wrapper(*args, **kwargs):
            settings = _update_settings_from_kwargs(
                backend.settings,
                issuers=issuers, required=required,
                subject_should_match_issuer=subject_should_match_issuer
            )

            request = None
            if len(args) > 0:
                request = args[0]

            error_response = _process_asap_token(
                request, backend, settings
            )

            if error_response is not None:
                return error_response

            return func(*args, **kwargs)

        return with_asap_wrapper

    if callable(func):
        return with_asap_decorator(func)

    return with_asap_decorator


def _restrict_asap(func=None, backend=None, issuers=None,
                   required=True, subject_should_match_issuer=None):
    """Decorator to allow endpoint-specific ASAP authorization, assuming ASAP
    authentication has already occurred.
    """

    def restrict_asap_decorator(func):
        @wraps(func)
        def restrict_asap_wrapper(request, *args, **kwargs):
            settings = _update_settings_from_kwargs(
                backend.settings,
                issuers=issuers, required=required,
                subject_should_match_issuer=subject_should_match_issuer
            )
            asap_claims = getattr(request, 'asap_claims', None)
            error_response = None

            if required and not asap_claims:
                return backend.get_401_response(
                    'Unauthorized', request=request
                )

            try:
                _verify_issuers(asap_claims, settings.ASAP_VALID_ISSUERS)
            except InvalidIssuerError:
                error_response = backend.get_403_response(
                    'Forbidden: Invalid token issuer', request=request
                )
            except InvalidTokenError:
                error_response = backend.get_401_response(
                    'Unauthorized: Invalid token', request=request
                )

            if error_response and required:
                return error_response

            return func(request, *args, **kwargs)

        return restrict_asap_wrapper

    if callable(func):
        return restrict_asap_decorator(func)

    return restrict_asap_decorator


def _update_settings_from_kwargs(settings, issuers=None, required=True,
                                 subject_should_match_issuer=None):
    settings = settings.copy()

    if issuers is not None:
        settings['ASAP_VALID_ISSUERS'] = set(issuers)

    if required is not None:
        settings['ASAP_REQUIRED'] = required

    if subject_should_match_issuer is not None:
        settings['ASAP_SUBJECT_SHOULD_MATCH_ISSUER'] = (
            subject_should_match_issuer
        )

    return SettingsDict(settings)



import warnings

from .decorators import requires_asap  # noqa


warnings.warn(
    "The atlassian_jwt_auth.contrib.flask_app package is deprecated in 4.0.0 "
    "in favour of atlassian_jwt_auth.frameworks.flask.",
    DeprecationWarning, stacklevel=2
)



import os
import shutil
import tempfile
import unittest


import atlassian_jwt_auth
from atlassian_jwt_auth import key
from atlassian_jwt_auth.tests import utils


class BaseJWTAuthSignerWithFilePrivateKeyRetrieverTest(object):

    """ tests for the JWTAuthSigner using the FilePrivateKeyRetriever. """

    def setUp(self):
        self.test_dir = tempfile.mkdtemp(prefix='atlassian-jwt-p-tests')
        self.key_dir = os.path.join(self.test_dir, 'jwtprivatekeys')
        for dir in ['invalid-issuer', 'issuer-with-many-keys',
                    'valid-issuer']:
            os.makedirs(os.path.join(self.key_dir, dir))
        self._private_key_pem = self.get_new_private_key_in_pem_format()
        for file_loc in [
            'invalid-issuer/key-tests-pem.new',
            'issuer-with-many-keys/key1.pem.new',
            'issuer-with-many-keys/key2.pem',
            'issuer-with-many-keys/key3.pem',
            'issuer-with-many-keys/key4.pem.new',
            'valid-issuer/key-for-tests.pem'
        ]:
            file_location = os.path.join(self.key_dir, file_loc)
            with open(file_location, 'wb') as f:
                f.write(self._private_key_pem)

    def tearDown(self):
        if self.test_dir:
            shutil.rmtree(self.test_dir)

    def create_signer_for_issuer(self, issuer):
        return \
            atlassian_jwt_auth.create_signer_from_file_private_key_repository(
                issuer, self.key_dir, algorithm=self.algorithm)

    def test_succeeds_if_issuer_has_one_valid_key(self):
        signer = self.create_signer_for_issuer('valid-issuer')
        token = signer.generate_jwt('audience')
        self.assertIsNotNone(token)

    def test_picks_last_valid_key_id(self):
        signer = self.create_signer_for_issuer('issuer-with-many-keys')
        token = signer.generate_jwt('audience')
        key_identifier = key._get_key_id_from_jwt_header(token)
        expected_key_id = 'issuer-with-many-keys/key3.pem'
        self.assertEqual(key_identifier.key_id, expected_key_id)

    def test_fails_if_issuer_has_no_valid_keys(self):
        signer = self.create_signer_for_issuer('invalid-issuer')
        with self.assertRaisesRegex(IOError, 'Issuer has no valid keys'):
            signer.generate_jwt('audience')

    def test_fails_if_issuer_does_not_exist(self):
        signer = self.create_signer_for_issuer('this-does-not-exist')
        with self.assertRaisesRegex(OSError, 'No such file or directory'):
            signer.generate_jwt('audience')


class JWTAuthSignerWithFilePrivateKeyRetrieverRS256Test(
        BaseJWTAuthSignerWithFilePrivateKeyRetrieverTest,
        utils.RS256KeyTestMixin,
        unittest.TestCase):
    pass


class JWTAuthSignerWithFilePrivateKeyRetrieverES256Test(
        BaseJWTAuthSignerWithFilePrivateKeyRetrieverTest,
        utils.ES256KeyTestMixin,
        unittest.TestCase):
    pass



from django.conf.urls import url

from atlassian_jwt_auth.frameworks.django.tests import views


urlpatterns = [
    url(r'^asap/expected$', views.expected_view, name='expected'),
    url(r'^asap/unexpected$', views.unexpected_view, name='unexpected'),
    url(r'^asap/decorated$', views.decorated_view, name='decorated'),
    url(r'^asap/settings$', views.settings_view, name='settings'),

    url(r'^asap/subject_does_not_need_to_match_issuer$',
        views.subject_does_not_need_to_match_issuer_view,
        name='subject_does_not_need_to_match_issuer'),
    url(r'^asap/subject_does_need_to_match_issuer_view$',
        views.subject_does_need_to_match_issuer_view,
        name='subject_does_need_to_match_issuer'),

    url(r'^asap/subject_does_not_need_to_match_issuer_from_settings$',
        views.subject_does_not_need_to_match_issuer_from_settings_view,
        name='subject_does_not_need_to_match_issuer_from_settings'),

    url(r'^asap/needed$', views.needed_view, name='needed'),
    url(r'^asap/unneeded$', views.unneeded_view, name='unneeded'),
    url(r'^asap/restricted_issuer$', views.restricted_issuer_view,
        name='restricted_issuer'),
    url(r'^asap/restricted_subject$', views.restricted_subject_view,
        name='restricted_subject'),
]



import unittest
from datetime import timedelta

import jwt
from requests import Request

import atlassian_jwt_auth
from atlassian_jwt_auth.tests import utils
from atlassian_jwt_auth.contrib.requests import JWTAuth, create_jwt_auth


class BaseRequestsTest(object):

    """ tests for the contrib.requests.JWTAuth class """
    auth_cls = JWTAuth

    def setUp(self):
        self._private_key_pem = self.get_new_private_key_in_pem_format()
        self._public_key_pem = utils.get_public_key_pem_for_private_key_pem(
            self._private_key_pem)

    def assert_authorization_header_is_valid(self, auth):
        """ asserts that the given request contains a valid Authorization
            header.
        """
        auth_header = self._get_auth_header(auth)
        bearer = auth_header.split(b' ')[1]
        # Decode the JWT (verifying the signature and aud match)
        # an exception is thrown if this fails
        algorithms = atlassian_jwt_auth.get_permitted_algorithm_names()
        return jwt.decode(bearer, self._public_key_pem.decode(),
                          audience='audience', algorithms=algorithms)

    def _get_auth_header(self, auth):
        request = auth(Request())
        auth_header = request.headers['Authorization']
        return auth_header

    def create_jwt_auth(self, *args, **kwargs):
        return create_jwt_auth(*args, **kwargs)

    def test_JWTAuth_make_authenticated_request(self):
        """Verify a valid Authorization header is added by JWTAuth"""
        jwt_auth_signer = atlassian_jwt_auth.create_signer(
            'issuer',
            'issuer/key',
            self._private_key_pem.decode(),
            algorithm=self.algorithm)
        auth = self.auth_cls(jwt_auth_signer, 'audience')
        self.assert_authorization_header_is_valid(auth)

    def test_create_jwt_auth(self):
        """Verify a valid Authorization header is added by JWTAuth"""
        auth = self.create_jwt_auth('issuer', 'issuer/key',
                                    self._private_key_pem.decode(), 'audience',
                                    algorithm=self.algorithm)
        self.assert_authorization_header_is_valid(auth)

    def test_create_jwt_auth_with_additional_claims(self):
        """ Verify a Valid Authorization header is added by JWTAuth and
            contains the additional claims when provided.
        """
        jwt_auth_signer = atlassian_jwt_auth.create_signer(
            'issuer',
            'issuer/key',
            self._private_key_pem.decode(),
            algorithm=self.algorithm)
        auth = self.auth_cls(jwt_auth_signer, 'audience',
                             additional_claims={'example': 'claim'})
        token = self.assert_authorization_header_is_valid(auth)
        self.assertEqual(token.get('example'), 'claim')

    def test_do_not_reuse_jwts(self):
        auth = self.create_jwt_auth('issuer', 'issuer/key',
                                    self._private_key_pem.decode(), 'audience',
                                    algorithm=self.algorithm)
        auth_header = self._get_auth_header(auth)
        self.assertNotEqual(auth_header, self._get_auth_header(auth))

    def test_reuse_jwts(self):
        auth = self.create_jwt_auth('issuer', 'issuer/key',
                                    self._private_key_pem.decode(), 'audience',
                                    algorithm=self.algorithm, reuse_jwts=True)
        auth_header = self._get_auth_header(auth)
        self.assertEqual(auth_header, self._get_auth_header(auth))

    def test_do_not_reuse_jwt_if_audience_changes(self):
        auth = self.create_jwt_auth('issuer', 'issuer/key',
                                    self._private_key_pem.decode(), 'audience',
                                    algorithm=self.algorithm, reuse_jwts=True)
        auth_header = self._get_auth_header(auth)
        auth._audience = 'not-' + auth._audience
        self.assertNotEqual(auth_header, self._get_auth_header(auth))

    def test_do_not_reuse_jwt_if_issuer_changes(self):
        auth = self.create_jwt_auth('issuer', 'issuer/key',
                                    self._private_key_pem.decode(), 'audience',
                                    algorithm=self.algorithm, reuse_jwts=True)
        auth_header = self._get_auth_header(auth)
        auth._signer.issuer = 'not-' + auth._signer.issuer
        self.assertNotEqual(auth_header, self._get_auth_header(auth))

    def test_do_not_reuse_jwt_if_lifetime_changes(self):
        auth = self.create_jwt_auth('issuer', 'issuer/key',
                                    self._private_key_pem.decode(), 'audience',
                                    algorithm=self.algorithm, reuse_jwts=True)
        auth_header = self._get_auth_header(auth)
        auth._signer.lifetime = auth._signer.lifetime - timedelta(seconds=1)
        self.assertNotEqual(auth_header, self._get_auth_header(auth))

    def test_do_not_reuse_jwt_if_subject_changes(self):
        auth = self.create_jwt_auth('issuer', 'issuer/key',
                                    self._private_key_pem.decode(), 'audience',
                                    algorithm=self.algorithm, reuse_jwts=True,
                                    subject='subject')
        auth_header = self._get_auth_header(auth)
        auth._signer.subject = 'not-' + auth._signer.subject
        self.assertNotEqual(auth_header, self._get_auth_header(auth))

    def test_do_not_reuse_jwt_if_additional_claims_change(self):
        auth = self.create_jwt_auth('issuer', 'issuer/key',
                                    self._private_key_pem.decode(), 'audience',
                                    algorithm=self.algorithm, reuse_jwts=True)
        auth_header = self._get_auth_header(auth)
        auth._additional_claims['foo'] = 'bar'
        self.assertNotEqual(auth_header, self._get_auth_header(auth))

    def test_reuse_jwt_with_additional_claims(self):
        # calculating the cache key with additional claims is non-trivial
        auth = self.create_jwt_auth('issuer', 'issuer/key',
                                    self._private_key_pem.decode(), 'audience',
                                    algorithm=self.algorithm, reuse_jwts=True)
        auth._additional_claims['foo'] = 'bar'
        auth._additional_claims['fool'] = 'blah'
        auth._additional_claims['foot'] = 'quux'
        auth_header = self._get_auth_header(auth)
        self.assertEqual(auth_header, self._get_auth_header(auth))


class RequestsRS256Test(BaseRequestsTest,
                        utils.RS256KeyTestMixin,
                        unittest.TestCase):
    pass


class RequestsES256Test(BaseRequestsTest,
                        utils.ES256KeyTestMixin,
                        unittest.TestCase):
    pass



from collections import namedtuple
from ..common.asap import _process_asap_token
from .backend import WSGIBackend

Request = namedtuple('Request', ['environ', 'start_response'])


class ASAPMiddleware(object):
    def __init__(self, handler, settings):
        self._next = handler
        self._backend = WSGIBackend(settings)
        self._verifier = self._backend.get_verifier()

    def __call__(self, environ, start_response):
        settings = self._backend.settings
        request = Request(environ, start_response)
        error_response = _process_asap_token(
            request, self._backend, settings, verifier=self._verifier
        )
        if error_response is not None:
            return error_response

        return self._next(environ, start_response)



from atlassian_jwt_auth.algorithms import get_permitted_algorithm_names  # noqa

from atlassian_jwt_auth.signer import (  # noqa
    create_signer,
    create_signer_from_file_private_key_repository,
)

from atlassian_jwt_auth.key import (  # noqa
    KeyIdentifier,
    HTTPSPublicKeyRetriever,
)

from atlassian_jwt_auth.verifier import (  # noqa
    JWTAuthVerifier,
)



import os

import django
from django.test.testcases import SimpleTestCase
from django.test.utils import override_settings, modify_settings

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

from atlassian_jwt_auth.contrib.tests.utils import (
    get_static_retriever_class,
)
from atlassian_jwt_auth.tests import utils
from atlassian_jwt_auth.tests.utils import (
    create_token,
    RS256KeyTestMixin,
)


class DjangoAsapMixin(object):

    @classmethod
    def setUpClass(cls):
        os.environ.setdefault(
            'DJANGO_SETTINGS_MODULE',
            'atlassian_jwt_auth.frameworks.django.tests.settings')

        django.setup()
        super(DjangoAsapMixin, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        super(DjangoAsapMixin, cls).tearDownClass()
        del os.environ['DJANGO_SETTINGS_MODULE']

    def setUp(self):
        super(DjangoAsapMixin, self).setUp()
        self._private_key_pem = self.get_new_private_key_in_pem_format()
        self._public_key_pem = utils.get_public_key_pem_for_private_key_pem(
            self._private_key_pem
        )

        self.retriever = get_static_retriever_class({
            'client-app/key01': self._public_key_pem
        })

        self.test_settings = {
            'ASAP_KEY_RETRIEVER_CLASS': self.retriever
        }


@modify_settings(MIDDLEWARE={
    'prepend': 'atlassian_jwt_auth.frameworks.django.asap_middleware',
})
class TestAsapMiddleware(DjangoAsapMixin, RS256KeyTestMixin, SimpleTestCase):

    def check_response(self,
                       view_name,
                       response_content='',
                       status_code=200,
                       issuer='client-app',
                       audience='server-app',
                       key_id='client-app/key01',
                       subject=None,
                       private_key=None,
                       token=None,
                       authorization=None,
                       retriever_key=None):
        if authorization is None:
            if token is None:
                if private_key is None:
                    private_key = self._private_key_pem
                token = create_token(issuer=issuer, audience=audience,
                                     key_id=key_id, private_key=private_key,
                                     subject=subject)
            authorization = b'Bearer ' + token

        test_settings = self.test_settings.copy()
        if retriever_key is not None:
            retriever = get_static_retriever_class({
                retriever_key: self._public_key_pem
            })
            test_settings['ASAP_KEY_RETRIEVER_CLASS'] = retriever

        with override_settings(**test_settings):
            response = self.client.get(reverse(view_name),
                                       HTTP_AUTHORIZATION=authorization)

        self.assertContains(response, response_content,
                            status_code=status_code)

    def test_request_with_valid_token_is_allowed(self):
        self.check_response('needed', 'one', 200)

    def test_request_with_duplicate_jti_is_rejected_as_per_setting(self):
        self.test_settings['ASAP_CHECK_JTI_UNIQUENESS'] = True
        token = create_token(
            issuer='client-app', audience='server-app',
            key_id='client-app/key01', private_key=self._private_key_pem
        )
        str_auth = 'Bearer ' + token.decode(encoding='iso-8859-1')
        self.check_response('needed', 'one', 200, authorization=str_auth)
        self.check_response('needed', 'duplicate jti', 401,
                            authorization=str_auth)

    def _assert_request_with_duplicate_jti_is_accepted(self):
        token = create_token(
            issuer='client-app', audience='server-app',
            key_id='client-app/key01', private_key=self._private_key_pem
        )
        str_auth = 'Bearer ' + token.decode(encoding='iso-8859-1')
        self.check_response('needed', 'one', 200, authorization=str_auth)
        self.check_response('needed', 'one', 200, authorization=str_auth)

    def test_request_with_duplicate_jti_is_accepted(self):
        self._assert_request_with_duplicate_jti_is_accepted()

    def test_request_with_duplicate_jti_is_accepted_as_per_setting(self):
        self.test_settings['ASAP_CHECK_JTI_UNIQUENESS'] = False
        self._assert_request_with_duplicate_jti_is_accepted()

    def test_request_with_string_headers_is_allowed(self):
        token = create_token(
            issuer='client-app', audience='server-app',
            key_id='client-app/key01', private_key=self._private_key_pem
        )
        str_auth = 'Bearer ' + token.decode(encoding='iso-8859-1')
        self.check_response('needed', 'one', 200, authorization=str_auth)

    def test_request_with_invalid_audience_is_rejected(self):
        self.check_response('needed', 'Unauthorized', 401,
                            audience='invalid')

    def test_request_with_invalid_token_is_rejected(self):
        self.check_response('needed', 'Unauthorized', 401,
                            authorization='Bearer invalid')

    def test_request_without_token_is_rejected(self):
        with override_settings(**self.test_settings):
            response = self.client.get(reverse('needed'))

        self.assertContains(response, 'Unauthorized',
                            status_code=401)

    def test_request_with_invalid_issuer_is_rejected(self):
        self.check_response('needed', 'Forbidden', 403,
                            issuer='something-invalid',
                            key_id='something-invalid/key01',
                            retriever_key='something-invalid/key01')

    def test_request_non_whitelisted_decorated_issuer_is_rejected(self):
        self.check_response('needed', 'Forbidden', 403,
                            issuer='unexpected',
                            key_id='unexpected/key01',
                            retriever_key='unexpected/key01')

    def test_request_non_decorated_issuer_is_rejected(self):
        self.check_response('restricted_issuer', 'Forbidden', 403)

    def test_request_decorated_issuer_is_allowed(self):
        self.check_response('restricted_issuer', 'three',
                            issuer='whitelist',
                            key_id='whitelist/key01',
                            retriever_key='whitelist/key01')

    # TODO: modify JWTAuthSigner to allow non-issuer subjects and update the
    # decorated subject test cases
    def test_request_non_decorated_subject_is_rejected(self):
        self.check_response('restricted_subject', 'Forbidden', 403,
                            issuer='whitelist',
                            key_id='whitelist/key01',
                            retriever_key='whitelist/key01')

    def test_request_using_settings_only_is_allowed(self):
        self.check_response('unneeded', 'two')

    def test_request_subject_does_not_need_to_match_issuer_from_settings(self):
        self.test_settings['ASAP_SUBJECT_SHOULD_MATCH_ISSUER'] = False
        self.check_response('needed', 'one', 200, subject='different_than_is')

    def test_request_subject_and_issue_not_matching(self):
        self.check_response(
            'needed',
            'Subject and Issuer do not match',
            401,
            subject='different_than_is',
        )


class TestAsapDecorator(DjangoAsapMixin, RS256KeyTestMixin, SimpleTestCase):
    def test_request_with_valid_token_is_allowed(self):
        token = create_token(
            issuer='client-app', audience='server-app',
            key_id='client-app/key01', private_key=self._private_key_pem
        )
        with override_settings(**self.test_settings):
            response = self.client.get(reverse('expected'),
                                       HTTP_AUTHORIZATION=b'Bearer ' + token)

        self.assertContains(response, 'Greatest Success!', status_code=200)

    def test_request_with_string_headers_is_allowed(self):
        token = create_token(
            issuer='client-app', audience='server-app',
            key_id='client-app/key01', private_key=self._private_key_pem
        )
        str_token = token.decode(encoding='iso-8859-1')
        with override_settings(**self.test_settings):
            response = self.client.get(reverse('expected'),
                                       HTTP_AUTHORIZATION='Bearer ' +
                                                          str_token)

        self.assertContains(response, 'Greatest Success!', status_code=200)

    def test_request_with_invalid_audience_is_rejected(self):
        token = create_token(
            issuer='client-app', audience='something-invalid',
            key_id='client-app/key01', private_key=self._private_key_pem
        )
        with override_settings(**self.test_settings):
            response = self.client.get(reverse('expected'),
                                       HTTP_AUTHORIZATION=b'Bearer ' + token)

        self.assertContains(response, 'Unauthorized: Invalid token',
                            status_code=401)

    def test_request_with_invalid_token_is_rejected(self):
        with override_settings(**self.test_settings):
            response = self.client.get(
                reverse('expected'),
                HTTP_AUTHORIZATION=b'Bearer notavalidtoken')

        self.assertContains(response, 'Unauthorized: Invalid token',
                            status_code=401)

    def test_request_without_token_is_rejected(self):
        with override_settings(**self.test_settings):
            response = self.client.get(reverse('expected'))

        self.assertContains(response, 'Unauthorized',
                            status_code=401)

    def test_request_with_invalid_issuer_is_rejected(self):
        retriever = get_static_retriever_class({
            'something-invalid/key01': self._public_key_pem
        })
        token = create_token(
            issuer='something-invalid', audience='server-app',
            key_id='something-invalid/key01', private_key=self._private_key_pem
        )
        with override_settings(ASAP_KEY_RETRIEVER_CLASS=retriever):
            response = self.client.get(reverse('expected'),
                                       HTTP_AUTHORIZATION=b'Bearer ' + token)

        self.assertContains(response, 'Forbidden: Invalid token issuer',
                            status_code=403)

    def test_request_non_decorated_issuer_is_rejected(self):
        token = create_token(
            issuer='client-app', audience='server-app',
            key_id='client-app/key01', private_key=self._private_key_pem
        )
        with override_settings(**self.test_settings):
            response = self.client.get(reverse('decorated'),
                                       HTTP_AUTHORIZATION=b'Bearer ' + token)

        self.assertContains(response, 'Forbidden: Invalid token issuer',
                            status_code=403)

    def test_request_decorated_issuer_is_allowed(self):
        retriever = get_static_retriever_class({
            'whitelist/key01': self._public_key_pem
        })
        token = create_token(
            issuer='whitelist', audience='server-app',
            key_id='whitelist/key01', private_key=self._private_key_pem
        )
        with override_settings(ASAP_KEY_RETRIEVER_CLASS=retriever):
            response = self.client.get(reverse('decorated'),
                                       HTTP_AUTHORIZATION=b'Bearer ' + token)

        self.assertContains(response, 'Only the right issuer is allowed.')

    def test_request_using_settings_only_is_allowed(self):
        token = create_token(
            issuer='client-app', audience='server-app',
            key_id='client-app/key01', private_key=self._private_key_pem
        )
        with override_settings(**self.test_settings):
            response = self.client.get(reverse('settings'),
                                       HTTP_AUTHORIZATION=b'Bearer ' + token)

        self.assertContains(response, 'Any settings issuer is allowed.')

    def test_request_subject_does_not_need_to_match_issuer(self):
        token = create_token(
            issuer='client-app', audience='server-app',
            key_id='client-app/key01', private_key=self._private_key_pem,
            subject='not-client-app',
        )
        with override_settings(**self.test_settings):
            response = self.client.get(
                reverse('subject_does_not_need_to_match_issuer'),
                HTTP_AUTHORIZATION=b'Bearer ' + token)

        self.assertContains(response, 'Subject does not need to match issuer.')

    def test_request_subject_does_need_to_match_issuer_override_settings(self):
        """ tests that the with_asap decorator can override the
            ASAP_SUBJECT_SHOULD_MATCH_ISSUER setting.
        """
        token = create_token(
            issuer='client-app', audience='server-app',
            key_id='client-app/key01', private_key=self._private_key_pem,
            subject='not-client-app',
        )
        with override_settings(**dict(
                self.test_settings, ASAP_SUBJECT_SHOULD_MATCH_ISSUER=False)):
            response = self.client.get(
                reverse('subject_does_need_to_match_issuer'),
                HTTP_AUTHORIZATION=b'Bearer ' + token)
            self.assertContains(
                response,
                'Unauthorized: Subject and Issuer do not match',
                status_code=401
            )

    def test_request_subject_does_not_need_to_match_issuer_from_settings(self):
        token = create_token(
            issuer='client-app', audience='server-app',
            key_id='client-app/key01', private_key=self._private_key_pem,
            subject='not-client-app',
        )
        with override_settings(**dict(
                self.test_settings, ASAP_SUBJECT_SHOULD_MATCH_ISSUER=False)):
            response = self.client.get(
                reverse('subject_does_not_need_to_match_issuer_from_settings'),
                HTTP_AUTHORIZATION=b'Bearer ' + token)

        self.assertContains(
            response, 'Subject does not need to match issuer (settings).')



import unittest

import atlassian_jwt_auth


class TestKeyModule(unittest.TestCase):

    """ tests for the key module. """

    def test_key_identifier_with_invalid_keys(self):
        """ test that invalid key identifiers are not permitted. """
        keys = ['../aha', '/a', r'\c:a', 'lk2j34/#$', 'a../../a', 'a/;a',
                ' ', ' / ', ' /',
                u'dir/some\0thing', 'a/#a', 'a/a?x', 'a/a;',
                ]
        for key in keys:
            with self.assertRaises(ValueError):
                atlassian_jwt_auth.KeyIdentifier(identifier=key)

    def test_key_identifier_with_valid_keys(self):
        """ test that valid keys work as expected. """
        for key in ['oa.oo/a', 'oo.sasdf.asdf/yes', 'oo/o']:
            key_id = atlassian_jwt_auth.KeyIdentifier(identifier=key)
            self.assertEqual(key_id.key_id, key)



import logging

from jwt.exceptions import InvalidIssuerError, InvalidTokenError

from atlassian_jwt_auth.exceptions import (
    PublicKeyRetrieverException,
    NoTokenProvidedError,
    JtiUniquenessException,
    SubjectDoesNotMatchIssuerException,
)


def _process_asap_token(request, backend, settings, verifier=None):
    """ Verifies an ASAP token, validates the claims, and returns an error
    response"""
    logger = logging.getLogger('asap')
    token = backend.get_asap_token(request)
    error_response = None
    if token is None and not settings.ASAP_REQUIRED and (
            settings.ASAP_REQUIRED is not None):
        return
    try:
        if token is None:
            raise NoTokenProvidedError
        if verifier is None:
            verifier = backend.get_verifier(settings=settings)
        asap_claims = verifier.verify_jwt(
            token,
            settings.ASAP_VALID_AUDIENCE,
            leeway=settings.ASAP_VALID_LEEWAY,
        )

        _verify_issuers(asap_claims, settings.ASAP_VALID_ISSUERS)
        backend.set_asap_claims_for_request(request, asap_claims)
    except NoTokenProvidedError:
        logger.info('No token provided')
        error_response = backend.get_401_response(
            'Unauthorized', request=request
        )
    except PublicKeyRetrieverException as e:
        if e.status_code not in (403, 404):
            # Any error other than "not found" is a problem and should
            # be dealt with elsewhere.
            # Note that we treat 403 like 404 to account for the fact
            # that a server configured to secure directory listings
            # will return 403 for a missing file to avoid leaking
            # information.
            raise
            logger.warning('Could not retrieve the matching public key')
        error_response = backend.get_401_response(
            'Unauthorized: Key not found', request=request
        )
    except InvalidIssuerError:
        logger.warning('Invalid token - issuer')
        error_response = backend.get_403_response(
            'Forbidden: Invalid token issuer', request=request
        )
    except InvalidTokenError:
        logger.warning('Invalid token')
        error_response = backend.get_401_response(
            'Unauthorized: Invalid token', request=request
        )
    except JtiUniquenessException:
        logger.warning('Invalid token - duplicate jti')
        error_response = backend.get_401_response(
            'Unauthorized: Invalid token - duplicate jti', request=request
        )
    except SubjectDoesNotMatchIssuerException:
        logger.warning('Invalid token - subject and issuer do not match')
        error_response = backend.get_401_response(
            'Unauthorized: Subject and Issuer do not match', request=request
        )
    except Exception:
        logger.exception('An error occured while checking an asap token')
        raise

    if error_response is not None and settings.ASAP_REQUIRED:
        return error_response


def _verify_issuers(asap_claims, issuers=None):
    """Verify that the issuer in the claims is valid and is expected."""
    claim_iss = asap_claims.get('iss')
    if issuers and claim_iss not in issuers:
        raise InvalidIssuerError



from __future__ import absolute_import

import atlassian_jwt_auth


class BaseJWTAuth(object):
    """Adds a JWT bearer token to the request per the ASAP specification"""

    def __init__(self, signer, audience, *args, **kwargs):
        self._audience = audience
        self._signer = signer
        self._additional_claims = kwargs.get('additional_claims', {})

    @classmethod
    def create(cls, issuer, key_identifier, private_key_pem, audience,
               **kwargs):
        """Instantiate a JWTAuth while creating the signer inline"""
        signer = atlassian_jwt_auth.create_signer(issuer, key_identifier,
                                                  private_key_pem, **kwargs)
        return cls(signer, audience)

    def _get_header_value(self):
        return b'Bearer ' + self._signer.generate_jwt(
            self._audience, additional_claims=self._additional_claims)



"""Provide asyncio support"""
import sys

if sys.version_info >= (3, 5):
    try:
        import aiohttp  # noqa
        from .auth import JWTAuth  # noqa
        from .key import HTTPSPublicKeyRetriever  # noqa
        from .verifier import JWTAuthVerifier  # noqa
    except ImportError as e:
        import warnings
        warnings.warn(str(e))


del sys



import unittest

from atlassian_jwt_auth.contrib.aiohttp.auth import create_jwt_auth, JWTAuth
from atlassian_jwt_auth.tests import utils
from atlassian_jwt_auth.contrib.tests import test_requests


class BaseAuthTest(test_requests.BaseRequestsTest):
    """ tests for the contrib.aiohttp.JWTAuth class """
    auth_cls = JWTAuth

    def _get_auth_header(self, auth):
        return auth.encode().encode('latin1')

    def create_jwt_auth(self, *args, **kwargs):
        return create_jwt_auth(*args, **kwargs)


class RequestsRS256Test(BaseAuthTest,
                        utils.RS256KeyTestMixin,
                        unittest.TestCase):
    pass


class RequestsES256Test(BaseAuthTest,
                        utils.ES256KeyTestMixin,
                        unittest.TestCase):
    pass



from __future__ import absolute_import

from atlassian_jwt_auth.auth import BaseJWTAuth

from requests.auth import AuthBase


class JWTAuth(AuthBase, BaseJWTAuth):
    """Adds a JWT bearer token to the request per the ASAP specification"""

    def __call__(self, r):
        r.headers['Authorization'] = self._get_header_value()
        return r


def create_jwt_auth(
        issuer, key_identifier, private_key_pem, audience, **kwargs):
    """Instantiate a JWTAuth while creating the signer inline"""
    return JWTAuth.create(
        issuer, key_identifier, private_key_pem, audience, **kwargs)



from django.conf import settings
from django.utils.deprecation import MiddlewareMixin

from atlassian_jwt_auth.frameworks.django.middleware import (
    OldStyleASAPMiddleware
)


class ProxiedAsapMiddleware(OldStyleASAPMiddleware, MiddlewareMixin):
    """Enable client auth for ASAP-enabled services that are forwarding
    non-ASAP client requests.

    This must come before any authentication middleware."""

    def __init__(self, get_response=None):
        super(ProxiedAsapMiddleware, self).__init__()
        self.get_response = get_response

        # Rely on this header to tell us if a request has been forwarded
        # from an ASAP-enabled service; will overwrite X-Forwarded-For
        self.xfwd = getattr(settings, 'ASAP_PROXIED_FORWARDED_FOR_HEADER',
                            'HTTP_X_ASAP_FORWARDED_FOR')

        # This header won't always be set, i.e. some users will be anonymous
        self.xauth = getattr(settings, 'ASAP_PROXIED_AUTHORIZATION_HEADER',
                             'HTTP_X_ASAP_AUTHORIZATION')

    def process_request(self, request):
        error_response = super(ProxiedAsapMiddleware, self).process_request(
            request
        )

        if error_response:
            return error_response

        forwarded_for = request.META.pop(self.xfwd, None)
        if forwarded_for is None:
            return

        request.asap_forwarded = True
        request.META['HTTP_X_FORWARDED_FOR'] = forwarded_for

        asap_auth = request.META.pop('HTTP_AUTHORIZATION', None)
        orig_auth = request.META.pop(self.xauth, None)

        # Swap original client header in to allow regular auth middleware
        if orig_auth is not None:
            request.META['HTTP_AUTHORIZATION'] = orig_auth
        if asap_auth is not None:
            request.META[self.xauth] = asap_auth

    def process_view(self, request, view_func, view_args, view_kwargs):
        if not hasattr(request, 'asap_forwarded'):
            return

        # swap headers back into place
        asap_auth = request.META.pop(self.xauth, None)
        orig_auth = request.META.pop('HTTP_AUTHORIZATION', None)

        if asap_auth is not None:
            request.META['HTTP_AUTHORIZATION'] = asap_auth
        if orig_auth is not None:
            request.META[self.xauth] = orig_auth



from functools import wraps

from django.http.response import HttpResponse

from atlassian_jwt_auth.frameworks.django.decorators import with_asap


def validate_asap(issuers=None, subjects=None, required=True):
    """Decorator to allow endpoint-specific ASAP authorization, assuming ASAP
    authentication has already occurred.

    :param list issuers: A list of issuers that are allowed to use the
        endpoint.
    :param list subjects: A list of subjects that are allowed to use the
        endpoint.
    :param boolean required: Whether or not to require ASAP on this endpoint.
        Note that requirements will be still be verified if claims are present.
    """
    def validate_asap_decorator(func):
        @wraps(func)
        def validate_asap_wrapper(request, *args, **kwargs):
            asap_claims = getattr(request, 'asap_claims', None)
            if required and not asap_claims:
                message = 'Unauthorized: Invalid or missing token'
                response = HttpResponse(message, status=401)
                response['WWW-Authenticate'] = 'Bearer'
                return response

            if asap_claims:
                iss = asap_claims['iss']
                if issuers and iss not in issuers:
                    message = 'Forbidden: Invalid token issuer'
                    return HttpResponse(message, status=403)

                sub = asap_claims.get('sub')
                if subjects and sub not in subjects:
                    message = 'Forbidden: Invalid token subject'
                    return HttpResponse(message, status=403)

            return func(request, *args, **kwargs)

        return validate_asap_wrapper
    return validate_asap_decorator


def requires_asap(issuers=None, subject_should_match_issuer=None, func=None):
    """Decorator for Django endpoints to require ASAP

    :param list issuers: *required The 'iss' claims that this endpoint is from.
    """
    return with_asap(func=func,
                     required=True,
                     issuers=issuers,
                     subject_should_match_issuer=subject_should_match_issuer)



class _WrappedException(object):
    """Allow wrapping exceptions in a new class while preserving the original
    as an attribute.

    Note that while Python 2 and 3 both have reasonable ways to handle this,
    they're mutually incompatible. This is a simple, portable approach that
    should be sufficient for most use cases.
    """

    def __init__(self, *args, **kwargs):
        wrapped_args = [arg for arg in args]

        if args:
            orig = args[0]
            if isinstance(orig, Exception):

                wrapped_args[0] = str(orig)
                self.original_exception = getattr(orig, 'original_exception',
                                                  orig)
        super(_WrappedException, self).__init__(*wrapped_args, **kwargs)


class _WithStatus(object):
    """Allow an optional status_code attribute on wrapped exceptions.

    This should allow inspecting HTTP-related errors without having to know
    details about the HTTP client library.
    """

    def __init__(self, *args, **kwargs):
        status_code = kwargs.pop('status_code', None)
        super(_WithStatus, self).__init__(*args, **kwargs)
        self.status_code = status_code


class ASAPAuthenticationException(_WrappedException, ValueError):
    """Base class for exceptions raised by this library

    Inherits from ValueError to maintain backward compatibility
    with clients that caught ValueError previously.
    """


class PublicKeyRetrieverException(_WithStatus, ASAPAuthenticationException):
    """Raise when there are issues retrieving the public key"""


class PrivateKeyRetrieverException(_WithStatus, ASAPAuthenticationException):
    """Raise when there are issues retrieving the private key"""


class KeyIdentifierException(ASAPAuthenticationException):
    """Raise when there are issues validating the key identifier"""


class JtiUniquenessException(ASAPAuthenticationException):
    """Raise when a JTI is seen more than once. """


class SubjectDoesNotMatchIssuerException(ASAPAuthenticationException):
    """Raise when the subject and issuer differ. """


class NoTokenProvidedError(ASAPAuthenticationException):
    """Raise when no token is provided"""
    pass



from flask import Response, current_app, g, request as current_req

from ..common.backend import Backend


class FlaskBackend(Backend):
    def get_authorization_header(self, request=None):
        if request is None:
            request = current_req

        return request.headers.get('AUTHORIZATION', '')

    def get_401_response(self, data=None, headers=None, request=None):
        if headers is None:
            headers = {}

        headers.update(self.default_headers_401)

        return Response(data, status=401, headers=headers)

    def get_403_response(self, data=None, headers=None, request=None):
        return Response(data, status=403, headers=headers)

    def set_asap_claims_for_request(self, request, claims):
        g.asap_claims = claims

    @property
    def settings(self):
        settings = {}

        settings.update(self.default_settings)

        for k in settings.keys():
            value = current_app.config.get(k)
            if value is None:
                continue

            settings[k] = value

        return self._process_settings(settings)
